#include <iostream>
#include <thread>

#include <AudioToolbox/AudioToolbox.h>
#include <CoreAudio/CoreAudio.h>

#include <mach/mach_time.h>
#include <mach/mach.h>
#include <sys/mman.h>

#define BUFFER_SIZE 16

using namespace std;

double ClockFrequency()
{
   mach_timebase_info_data_t info;
   mach_timebase_info(&info);

   return 1e9 * info.denom / info.numer;
}

double Timer()
{
   static double freq(ClockFrequency());
   return mach_absolute_time() / freq;
}

class RecTest
{
   public:
      RecTest();
      ~RecTest();

      OSStatus CB(const AudioTimeStamp *inTimeStamp, UInt32 inNumberFrames);
      void ShowGlitchesLoop();

   private:
      OSStatus Init();

      AudioComponentInstance audioUnit_;
      double prevSampleTime_;
      std::atomic<int> glitches_;
};

static OSStatus recordingCallback(void *inRefCon,
                                  AudioUnitRenderActionFlags *ioActionFlags,
                                  const AudioTimeStamp *inTimeStamp,
                                  UInt32 inBusNumber,
                                  UInt32 inNumberFrames,
                                  AudioBufferList *ioData)
{
   return static_cast<RecTest *>(inRefCon)->CB(inTimeStamp, inNumberFrames);
}

RecTest::RecTest() :
   prevSampleTime_(0),
   glitches_(0)
{
   if (Init())
      throw std::runtime_error("zut");
}

RecTest::~RecTest()
{
   // Stop audio unit
   OSStatus status = AudioOutputUnitStop(audioUnit_);
   if (status != noErr)
      cerr << "Error stopping audio unit: " << status << endl;

   // Clean up
   status = AudioUnitUninitialize(audioUnit_);
   if (status != noErr)
      cerr << "Error uninitializing audio unit: " << status << endl;
}

OSStatus RecTest::CB(const AudioTimeStamp *inTimeStamp,
                     UInt32 inNumberFrames)
{
   if (prevSampleTime_ &&
       inTimeStamp->mSampleTime - prevSampleTime_ - inNumberFrames)
      ++glitches_;

   prevSampleTime_ = inTimeStamp->mSampleTime;

   return noErr;
}

void RecTest::ShowGlitchesLoop()
{
   int count(0);
   double startTime(Timer());

   while(true)
      if (glitches_ != count)
      {
         ++count;
         std::cout <<Timer()-startTime <<"\tglitched" <<std::endl;
      }
}

OSStatus RecTest::Init()
{
   // Open the default audio input unit
   AudioComponentDescription desc;
   desc.componentType = kAudioUnitType_Output;
   desc.componentSubType = kAudioUnitSubType_HALOutput;
   desc.componentManufacturer = kAudioUnitManufacturer_Apple;
   desc.componentFlags = 0;
   desc.componentFlagsMask = 0;

   AudioComponent inputComponent = AudioComponentFindNext(NULL, &desc);
   if (inputComponent == NULL) {
      cerr << "Error: Audio component not found" << endl;
      return -1;
   }

   OSStatus status = AudioComponentInstanceNew(inputComponent, &audioUnit_);
   if (status != noErr) {
      cerr << "Error creating audio component instance: " << status << endl;
      return status;
   }

   // Disable output
   UInt32 disableOutput = 0;
   status = AudioUnitSetProperty(audioUnit_,
                                 kAudioOutputUnitProperty_EnableIO,
                                 kAudioUnitScope_Output,
                                 0,
                                 &disableOutput,
                                 sizeof(disableOutput));
   if (status != noErr) {
      cerr << "Error disabling output: " << status << endl;
      return status;
   }

   // Set the input device
   AudioDeviceID dev(0);

   AudioObjectPropertyAddress addr =
      {kAudioHardwarePropertyDefaultInputDevice,
       kAudioObjectPropertyScopeGlobal,
       kAudioObjectPropertyElementMain};
   UInt32 size=sizeof(dev);
   status = AudioObjectGetPropertyData(
      kAudioObjectSystemObject, &addr, 0, 0, &size, &dev);
   if (status != noErr) {
      cerr << "Error getting default input device: " << status << endl;
      return status;
   }

   status = AudioUnitSetProperty(audioUnit_,
                                 kAudioOutputUnitProperty_CurrentDevice,
                                 kAudioUnitScope_Global,
                                 0, &dev, sizeof(dev));
   if (status != noErr) {
      cerr << "Error setting input device: " << status << endl;
      return status;
   }

   // Set buffer size
   {
      AudioObjectPropertyAddress addr =
         {kAudioDevicePropertyBufferFrameSize,
          kAudioDevicePropertyScopeInput, 0};

      UInt32 bufSize(BUFFER_SIZE);
      status = AudioObjectSetPropertyData(dev, &addr, 0, 0,
                                          sizeof(bufSize), &bufSize);
      if (status != noErr) {
         cerr << "Error setting buffer size: " << status << endl;
         return status;
      }
   }

   // Enable input
   UInt32 enableInput = 1;
   status = AudioUnitSetProperty(audioUnit_,
                                 kAudioOutputUnitProperty_EnableIO,
                                 kAudioUnitScope_Input,
                                 1,
                                 &enableInput,
                                 sizeof(enableInput));
   if (status != noErr) {
      cerr << "Error enabling input: " << status << endl;
      return status;
   }

   // Set input format
   AudioStreamBasicDescription audioFormat;
   audioFormat.mSampleRate = 44100;
   audioFormat.mFormatID = kAudioFormatLinearPCM;
   audioFormat.mFormatFlags = kAudioFormatFlagIsFloat | kAudioFormatFlagIsNonInterleaved;
   audioFormat.mFramesPerPacket = 1;
   audioFormat.mChannelsPerFrame = 1;
   audioFormat.mBitsPerChannel = sizeof(float) * 8;
   audioFormat.mBytesPerFrame = sizeof(float);
   audioFormat.mBytesPerPacket = sizeof(float);
   status = AudioUnitSetProperty(audioUnit_,
                                 kAudioUnitProperty_StreamFormat,
                                 kAudioUnitScope_Output,
                                 1,
                                 &audioFormat,
                                 sizeof(audioFormat));
   if (status != noErr) {
      cerr << "Error setting input format: " << status << endl;
      return status;
   }

   // Set recording callback
   AURenderCallbackStruct callbackStruct;
   callbackStruct.inputProc = recordingCallback;
   callbackStruct.inputProcRefCon = this;
   status = AudioUnitSetProperty(audioUnit_,
                                 kAudioOutputUnitProperty_SetInputCallback,
                                 kAudioUnitScope_Global,
                                 0,
                                 &callbackStruct,
                                 sizeof(callbackStruct));
   if (status != noErr) {
      cerr << "Error setting recording callback: " << status << endl;
      return status;
   }

   // Initialize and start audio unit
   status = AudioUnitInitialize(audioUnit_);
   if (status != noErr) {
      cerr << "Error initializing audio unit: " << status << endl;
      return status;
   }

   status = AudioOutputUnitStart(audioUnit_);
   if (status != noErr) {
      cerr << "Error starting audio unit: " << status << endl;
      return status;
   }

   return 0;
}

int main()
{
   // Use valloc to be memory page aligned.
   RecTest *rc(new(valloc(sizeof(RecTest))) RecTest);

   // rc is page aligned.
   if (mlock(&rc, sizeof(RecTest)))
   {
      cerr << "mlock() failed: " << strerror(errno) << endl;
      return EXIT_FAILURE;
   }

   cout << "Recording..." << endl;
   rc->ShowGlitchesLoop();
}
