#include <iostream>
#include <thread>

#include <CoreAudio/CoreAudio.h>
#include <AudioToolbox/AudioToolbox.h>

#include <sys/mman.h>

#include <mach/mach_time.h>
#include <mach/mach.h>

using namespace std;

std::atomic<void *> rtThreadStack;

static OSStatus recordingCallback(void *inRefCon,
                                  AudioUnitRenderActionFlags *ioActionFlags,
                                  const AudioTimeStamp *inTimeStamp,
                                  UInt32 inBusNumber,
                                  UInt32 inNumberFrames,
                                  AudioBufferList *ioData)
{
   rtThreadStack = &ioData;
   return noErr;
}

OSStatus Start()
{
   AudioComponentInstance audioUnit;

   // Open the default audio input unit
   AudioComponentDescription desc;
   desc.componentType = kAudioUnitType_Output;
   desc.componentSubType = kAudioUnitSubType_HALOutput;
   desc.componentManufacturer = kAudioUnitManufacturer_Apple;
   desc.componentFlags = 0;
   desc.componentFlagsMask = 0;

   AudioComponent inputComponent = AudioComponentFindNext(NULL, &desc);
   if (inputComponent == NULL) {
      cerr << "Error: Audio component not found" << endl;
      return -1;
   }

   OSStatus status = AudioComponentInstanceNew(inputComponent, &audioUnit);
   if (status != noErr) {
      cerr << "Error creating audio component instance: " << status << endl;
      return status;
   }

   // Disable output
   UInt32 disableOutput = 0;
   status = AudioUnitSetProperty(audioUnit,
                                 kAudioOutputUnitProperty_EnableIO,
                                 kAudioUnitScope_Output,
                                 0,
                                 &disableOutput,
                                 sizeof(disableOutput));
   if (status != noErr) {
      cerr << "Error disabling output: " << status << endl;
      return status;
   }

   // Set the input device
   AudioDeviceID dev(0);

   AudioObjectPropertyAddress addr =
      {kAudioHardwarePropertyDefaultInputDevice,
       kAudioObjectPropertyScopeGlobal,
       kAudioObjectPropertyElementMain};
   UInt32 size=sizeof(dev);
   status = AudioObjectGetPropertyData(
      kAudioObjectSystemObject, &addr, 0, 0, &size, &dev);
   if (status != noErr) {
      cerr << "Error getting default input device: " << status << endl;
      return status;
   }

   status = AudioUnitSetProperty(audioUnit,
                                 kAudioOutputUnitProperty_CurrentDevice,
                                 kAudioUnitScope_Global,
                                 0, &dev, sizeof(dev));
   if (status != noErr) {
      cerr << "Error setting input device: " << status << endl;
      return status;
   }

   // Enable input
   UInt32 enableInput = 1;
   status = AudioUnitSetProperty(audioUnit,
                                 kAudioOutputUnitProperty_EnableIO,
                                 kAudioUnitScope_Input,
                                 1,
                                 &enableInput,
                                 sizeof(enableInput));
   if (status != noErr) {
      cerr << "Error enabling input: " << status << endl;
      return status;
   }

   // Set input format
   AudioStreamBasicDescription audioFormat;
   audioFormat.mSampleRate = 44100;
   audioFormat.mFormatID = kAudioFormatLinearPCM;
   audioFormat.mFormatFlags = kAudioFormatFlagIsFloat | kAudioFormatFlagIsNonInterleaved;
   audioFormat.mFramesPerPacket = 1;
   audioFormat.mChannelsPerFrame = 1;
   audioFormat.mBitsPerChannel = sizeof(float) * 8;
   audioFormat.mBytesPerFrame = sizeof(float);
   audioFormat.mBytesPerPacket = sizeof(float);
   status = AudioUnitSetProperty(audioUnit,
                                 kAudioUnitProperty_StreamFormat,
                                 kAudioUnitScope_Output,
                                 1,
                                 &audioFormat,
                                 sizeof(audioFormat));
   if (status != noErr) {
      cerr << "Error setting input format: " << status << endl;
      return status;
   }

   // Set recording callback
   AURenderCallbackStruct callbackStruct;
   callbackStruct.inputProc = recordingCallback;
   callbackStruct.inputProcRefCon = 0;
   status = AudioUnitSetProperty(audioUnit,
                                 kAudioOutputUnitProperty_SetInputCallback,
                                 kAudioUnitScope_Global,
                                 0,
                                 &callbackStruct,
                                 sizeof(callbackStruct));
   if (status != noErr) {
      cerr << "Error setting recording callback: " << status << endl;
      return status;
   }

   // Initialize and start audio unit
   status = AudioUnitInitialize(audioUnit);
   if (status != noErr) {
      cerr << "Error initializing audio unit: " << status << endl;
      return status;
   }

   status = AudioOutputUnitStart(audioUnit);
   if (status != noErr) {
      cerr << "Error starting audio unit: " << status << endl;
      return status;
   }

   return 0;
}

int main()
{
   if (Start())
   {
      std::cout <<"failed starting audio" <<std::endl;
      return EXIT_FAILURE;
   }

   while (!rtThreadStack);

   std::cout <<"Checking real-time thread stack..." <<std::endl;

   bool wasIncore(true);

   while(true)
   {
      char incore;
      int ret = mincore(rtThreadStack, 1, &incore);
      if (ret)
      {
         std::cout <<"mincore failed: " <<strerror(errno) <<std::endl;
         break;
      }

      bool isIncore(incore & 1);

      if (wasIncore != isIncore)
      {
         if (isIncore)
            std::cout <<"in core" <<std::endl;
         else
            std::cout <<"not in core" <<std::endl;
      }

      wasIncore = isIncore;
   }
}
